# tags/tcl/tags-procs.tcl

ad_library {

    Procs to support site-wide tagging

    @author Michael Steigman (michael@steigman.net)
    @creation-date 2008-05-29
    @cvs-id $Id$
}

namespace eval ::tag {}
namespace eval ::tag::object {}
namespace eval ::tags::util {}
namespace eval ::tags::object {}
namespace eval ::tags::user {}
namespace eval ::tags::group {}
namespace eval ::tags::subsite {}
namespace eval ::tags::package {}

ad_proc -public ::tag::get_id {
    {-tag_text:required}
} {
    return tag_id or 0 if tag doesn't exists
} {
    return [db_string get_id { select tag_id from tags where tag_text = lower(:tag_text) } -default 0]
}

ad_proc -public ::tag::add {
    {-tag_text:required}
    {-locale "" }
    {-type "" }
} {
    add tag to system, return tag_id
} {

    # does tag exist?
    set tag_id [::tag::get_id -tag_text $tag_text]
    if { $tag_id } { return $tag_id }

    if {[empty_string_p $locale]} {
        set locale [ad_conn locale]
    }
    set tag_id [db_nextval acs_object_id_seq]
    db_dml insert_tag {
	insert into tags
	(tag_id, tag_text, locale)
	values
	(:tag_id, lower(:tag_text), :locale)
    }
    return $tag_id
}

ad_proc -public ::tag::delete {
    {-tag_id:required}
} {
    delete a tag
} {
    db_dml delete_tag { delete from tags where tag_id = :tag_id }
    return 1
}

ad_proc -private ::tags::util::parse {
    -string:required
} {
    tokenize a string using heuristics

    @returns list of tags (potentially lists as well)
} {
    set tags [list]
    if { [regexp "," $string] } {
	# assuming comma separated
	foreach tag [split $string ","] {
	    set tag [string trim $tag]
	    if { [string length $tag] > 0 } { lappend tags $tag }
	}
    } else {
	# assume one long tag
	lappend tags $string
    }
    return $tags
}

ad_proc -public ::tags::add {
    {-tag_list:required}
} {
    return ids for each tag in string, creating new tags as necesssary
} {
    set tag_ids [list]
    foreach tag [tags::util::parse -string $tag_list] {
	lappend tag_ids [tag::add -tag_text $tag]
    }
    return $tag_ids
}

ad_proc -public ::tags::get_ids {
    {-tag_list:required}
} {
    return ids for each tag in list
} {
    set tag_ids [list]
    foreach tag $tag_list {
	if { [tag::get_id -tag_text $tag] } {
	    lappend tag_ids [tag::get_id -tag_text $tag]
	}
    }
    return $tag_ids
}

ad_proc -public ::tag::object::add {
    {-object_id:required}
    {-user_id:required}
    {-tag:required}
} {
    add a single tag to object_id (for user_id)
} {
    # get current tag ids
    ::set current_tag_ids [::tags::get_ids -tag_list [get -object_id $object_id -user_id $user_id]]
    set tag_id [tag::add -tag_text $tag]
    if { [lsearch $current_tag_ids $tag_id] eq "-1" } {
	# will probably isolate this duplicate code (see below) in another proc
	db_dml tag_object {
	    insert into tagged_objects
	    (object_id, user_id, tag_id, tagged_on)
	    values
	    (:object_id, :user_id, :tag_id, current_timestamp)
	}
	# also update the tag's last_used column
	db_dml update_tag { update tags set last_used = current_timestamp where tag_id = :tag_id }
    }

}

ad_proc -public ::tag::object::remove {
    {-object_id:required}
    {-user_id:required}
    {-tag:required}
} {
    remove a single tag from an object_id (for user_id)
} {
    set tag_id [tag::get_id -tag_text $tag]
    if { $tag_id } {
	db_dml remove_tag { delete from tagged_objects
                             where tag_id = :tag_id
                               and user_id = :user_id
                               and object_id = :object_id }
    }
}

ad_proc -public ::tags::object::set {
    {-object_id:required}
    {-user_id:required}
    {-tag_list:required}
} {
    applies tag_list to object_id (for user_id). tag list can be empty.
} {
    # get old and new tags in list form
    ::set current_tag_ids [::tags::get_ids -tag_list [get -object_id $object_id -user_id $user_id]]
    if { $tag_list eq "" } {
	# zeroing out tag list. add proc expects and returns a list, potentially empty; don't want that
	::set new_tag_ids ""
    } else {
	# sorting by -unique removes any potential dupes in user-supplied list
	::set new_tag_ids [lsort -unique [::tags::add -tag_list $tag_list]]
    }

    # run through each new tag and add if it's not there
    foreach tag_id $new_tag_ids {
	if { [lsearch $current_tag_ids $tag_id] eq "-1" } {
	    db_dml tag_object {
		insert into tagged_objects
		(object_id, user_id, tag_id, tagged_on)
		values
		(:object_id, :user_id, :tag_id, current_timestamp)
	    }
	    # also update the tag's last_used column
	    db_dml update_tag { update tags set last_used = current_timestamp where tag_id = :tag_id }
	}
    }

    # run through each old tag and remove if it's not in new list
    foreach tag_id $current_tag_ids {
	if { [lsearch $new_tag_ids $tag_id] eq "-1" } {
	    db_dml untag_object {
		delete from tagged_objects
		 where object_id = :object_id
		   and user_id = :user_id
                   and tag_id = :tag_id
	    }
	}
    }
}

ad_proc -public ::tags::object::get {
    {-object_id:required}
    {-user_id ""}
} {
    get tags for an object
} {

    if { ![empty_string_p $user_id] } {
	::set user_clause "and user_id = :user_id"
    } else {
	::set user_clause ""
    }

    return [db_list select_object_tags "
	select distinct tag_text
	  from tagged_objects o join tags t on (o.tag_id = t.tag_id)
         where object_id = :object_id
               $user_clause
    "]
}

ad_proc -public ::tags::user::get {
    {-user_id:required}
    {-object_id ""}
} {
    get tags for a user
} {

    if { ![empty_string_p $object_id] } {
	::set object_clause "and object_id = :object_id"
    } else {
	::set object_clause ""
    }

    return [db_list select_user_tags "
	select distinct tag_text
	  from tagged_objects o join tags t on (o.tag_id = t.tag_id)
         where user_id = :user_id
               $object_clause
    "]
}

ad_proc -public ::tags::group::get {
    {-group_id:required}
} {
    get unique tags used by group of users
} {
    return [db_list select_group_tags {
	select distinct tag_text
	  from tagged_objects o join tags t on (o.tag_id = t.tag_id)
	where user_id in (select user_id
                            from group_approved_member_map
                           where group_id = :group_id
                             and rel_type = 'membership_rel')

    } ]
}

ad_proc -public ::tags::group::cloud {
    {-group_id:required}
} {
    generate cloud of tags used by a group of users
    (e.g. for members list filtering)
} {
    set tag_list [db_list_of_lists select_group_tags {
        select tag_text, count(*) as count
	  from tagged_objects o join tags t on (o.tag_id = t.tag_id)
         where user_id in (select user_id
                             from group_approved_member_map
                            where group_id = :group_id
                              and rel_type = 'membership_rel')
         group by tag_text
    } ]
    return [::tags::cloud -tag_list $tag_list]
}

ad_proc -public ::tags::subsite::get {
    {-subsite_id:required}
} {
    get tags used on objects created within an entire subsite
} {
    set subsite_node [subsite::get_element -subsite_id $subsite_id -element node_id]
    set tag_list [db_list select_subsite_tags "
	select distinct t.tag_text
	  from tagged_objects tob
          join acs_objects o on (tob.object_id = o.object_id)
          join tags t on (tob.tag_id = t.tag_id)
         where o.package_id in ([join [subsite::util::packages -node_id $subsite_node] ", "])
    "]
    return [lsort $tag_list]
}

ad_proc -public ::tags::subsite::cloud {
    {-subsite_id:required}
} {
    generate cloud with tags used on objects created within an entire subsite
} {
    set subsite_node [subsite::get_element -subsite_id $subsite_id -element node_id]
    set tag_list [db_list_of_lists select_subsite_tags_and_counts "
        select tag_text, count(*) as count
	  from tagged_objects tob
          join acs_objects o on (tob.object_id = o.object_id)
          join tags t on (tob.tag_id = t.tag_id)
         where o.package_id in ([join [subsite::util::packages -node_id $subsite_node] ", "])
         group by tag_text
    "]
    return [::tags::cloud -tag_list $tag_list]
}

ad_proc -public ::tags::package::get {
    {-package_id:required}
} {
    get tags used by objects in a package
} {
    set tag_list [db_list select_package_tags {
	select distinct t.tag_text
	  from tagged_objects tob
          join acs_objects o on (tob.object_id = o.object_id)
          join tags t on (tob.tag_id = t.tag_id)
         where o.package_id = :package_id
    }]
    return [lsort $tag_list]
}

ad_proc -public ::tags::package::cloud {
    {-package_id:required}
} {
    generate cloud of tags used by objects in a package
} {
    set tag_list [db_list_of_lists select_package_tags_and_counts {
        select tag_text, count(*) as count
	  from tagged_objects tob
          join acs_objects o on (tob.object_id = o.object_id)
          join tags t on (tob.tag_id = t.tag_id)
         where o.package_id = :package_id
         group by tag_text
    }]
    return [::tags::cloud -tag_list $tag_list]
}

ad_proc -private ::tags::cloud {
    {-steps {smallest smaller medium bigger biggest}}
    {-tag_list:required}
} {
    take list of lists containing tags and counts and return list of lists with
    tags and size which can be used in css
    (based on redwylie's 5 step css tag cloud or whatever is passed in to -steps)

    based on http://www.chasedavis.com/2007/jan/16/log-based-tag-clouds-python/
} {

    if { [llength $tag_list] == 0 } { return {} }

    # first, determine min/max weight
    set max_weight 0
    set min_weight [expr double([lindex [lindex $tag_list 0] 1])]
    foreach tag $tag_list {
        set tag_weight [lindex $tag 1]
        if {$tag_weight < $min_weight} {
            set min_weight [expr double($tag_weight)]
        }
        if {$tag_weight > $max_weight} {
            set max_weight [expr double($tag_weight)]
        }
    }

    # next, new delta
    set new_delta [expr [expr $max_weight - $min_weight] / [expr double([llength $steps])]]
    ns_log debug "_The Basics_ min $min_weight max $max_weight delta $new_delta"

    # now run through steps again and set css according to steps list
    set i 0
    set _tag_list [list]
    set _done [list]
    while { $i < [llength $steps] } {
	# set up threshold for this step
	set next_threshold [expr { 100.0 * log(($min_weight + ($i + 1) * $new_delta) + 2.0) }]
	ns_log debug "++ STEP $i THRESHOLD $next_threshold SIZE [lindex $steps $i]"
	foreach tag $tag_list {
	    # is the log of the tag's count less than our current threshold?
	    set tag_count [expr double([lindex $tag 1])]
	    if { [lsearch $_done [lindex $tag 0]] >= 0 } { set done_p 1 } else { set done_p 0 }
	    if { !$done_p } {
		ns_log debug "   =======> have TAG [lindex $tag 0] COUNT $tag_count SCORE [expr 100.0 * [expr log([expr [lindex $tag 1] + 2.0])]]"
	    }
	    if { [expr { 100.0 * log($tag_count + 2.0) } ] <= $next_threshold && !$done_p } {
		ns_log debug "          > doing TAG [lindex $tag 0]"
		# if we haven't already, set the css size here
		# (that would be the last stop for this tag)
		lappend _tag_list [list [lindex $tag 0] [lindex $steps $i]]
		# tcl 8.5 let's you search elements of sublists, which, when we have it, will obviate this helper list
		lappend _done [lindex $tag 0]
	    }
	}
	ns_log debug "++ END STEP $i DONE BUCKET $_done"
	ns_log debug "++ "
	incr i
    }
    return [lsort $_tag_list]
}
